/*
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://eu-egee.org/partners/ for details on the copyright holders.
 * For license conditions see the license file or http://eu-egee.org/license.html
 *
 * Authors:
 *     Gabor Gombas <Gabor.Gombas@cern.ch>
 *     Peter Kunszt <Peter.Kunszt@cern.ch>
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ServiceDiscoveryI.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <lber.h>
#include <ldap.h>
#include <glib.h>

#ifndef __GNUC__
#define __attribute__(x)
#endif

/**********************************************************************
 * Constants
 */

/* LDAP attributes */
#define ATTR_OC			"objectClass"
#define ATTR_DN			"dn"
#define ATTR_NAME		"GlueServiceUniqueID"
#define ATTR_TYPE		"GlueServiceType"
#define ATTR_VERSION		"GlueServiceVersion"
#define ATTR_LINK		"GlueForeignKey"
#define ATTR_SITE		"GlueSiteUniqueID"
#define ATTR_ENDPOINT		"GlueServiceEndpoint"
#define ATTR_VERSION		"GlueServiceVersion"
#define ATTR_STATUS             "GlueServiceStatus"
#define ATTR_WSDL		"GlueServiceWSDL"
#define ATTR_ADMIN
#define ATTR_VO			"GlueServiceOwner"
#define ATTR_DATAKEY		"GlueServiceDataKey"
#define ATTR_DATAVALUE		"GlueServiceDataValue"

// nonstandard glue stuff to fix LCG inconsistencies
#define ATTR_VERSION_MAJOR	"GlueServiceMajorVersion"
#define ATTR_VERSION_MINOR	"GlueServiceMinorVersion"
#define ATTR_VERSION_PATCH	"GlueServicePatchVersion"
#define ATTR_URI		"GlueServiceURI"
#define ATTR_NM			"GlueServiceName"
#define ATTR_ACCESSRULE         "GlueServiceAccessControlRule"
#define ATTR_HOSTINGORG         "GlueServiceHostingOrganization"
#define ATTR_SA_ROOT            "GlueSARoot"
#define ATTR_SA_ACBR            "GlueSAAccessControlBaseRule"
#define ATTR_SE	        	"GlueSEUniqueID"
#define ATTR_CHUNK		"GlueChunkKey"

/* LDAP object classes */
#define CLASS_SERVICE		"GlueService"
#define CLASS_DATA		"GlueServiceData"
#define CLASS_KEY		"GlueKey"
#define CLASS_SA		"GlueSA"

/* LDAP filter definitions */
#define QUERY_NAME		"(&" \
					"(" ATTR_OC "=" CLASS_SERVICE ")" \
					"(|(" ATTR_NAME "=%s)" \
					"(" ATTR_URI "=%s))" \
				")"

#define QUERY_DATA		"(" ATTR_OC "=" CLASS_DATA ")"

#define QUERY_DATAKEY		"(&" \
					QUERY_DATA \
					"(" ATTR_DATAKEY "=%s)" \
				")"

/* Query fragments for listServicesByData() */
#define QUERY_OC_SERVICE	"(" ATTR_OC "=" CLASS_SERVICE ")"
#define QUERY_OC_SA		"(" ATTR_OC "=" CLASS_SA ")"
#define QUERY_SITE		"(&" \
					"(" ATTR_OC "=" CLASS_KEY ")" \
					"(" ATTR_LINK "=" ATTR_SITE "=%s)" \
				")"
#define QUERY_TYPE		"(" ATTR_TYPE "=%s)"

#define QUERY_SRM_TYPE		"(|" \
					"(" ATTR_TYPE "=SRM)" \
					"(" ATTR_TYPE "=srm_v1)" \
				")"

#define QUERY_HOST		"(|(" ATTR_ENDPOINT "=*%s*)" \
                                      "(" ATTR_URI  "=*%s*))" 

#define QUERY_SE		"(" ATTR_CHUNK "=" ATTR_SE "=%s)"

/* Query expressing "No VO attribute" */
#define QUERY_VO_ANY		"(!(" ATTR_VO "=*))"

#define QUERY_VO_PRE		"(|" QUERY_VO_ANY
#define QUERY_VO		"(" ATTR_VO "=%s)"
#define QUERY_VO_POST		")"

/* XXX Possible values? */
#define QUERY_AVAILABLE		"(" ATTR_STATUS "=running)"

/* LDAP query timeout */
#define QUERY_TIMEOUT		60

/**********************************************************************
 * Prototypes
 */

static SDServiceDetails *sd_bdii_getServiceDetails(const char *name,
	SDException *exception);
static SDServiceList *sd_bdii_listAssociatedServices(const char *name,
	const char *type, const char *site, const SDVOList *vos,
	SDException *exception);
static SDServiceList *sd_bdii_listServicesByData(const SDServiceDataList *data,
	const char *type, const char *site, const SDVOList *vos,
	SDException *exception);
static SDServiceList *sd_bdii_listServicesByHost(const char *type, const char *host,
	const SDVOList *vos, SDException *exception);
static SDService *sd_bdii_getService(const char *name, SDException *exception);
static SDServiceDataList *sd_bdii_getServiceData(const char *name,
	SDException *exception);
static char *sd_bdii_getServiceDataItem(const char *serviceName, const char *key,
	SDException *exception);
static char *sd_bdii_getServiceSite(const char *serviceName, SDException *exception);
static char *sd_bdii_getServiceWSDL(const char *serviceName, SDException *exception);

static void sd_bdii_freeService(SDService *service);
static void sd_bdii_freeServiceList(SDServiceList *list);
static void sd_bdii_freeServiceDataList(SDServiceDataList *list);
static void sd_bdii_freeServiceDetails(SDServiceDetails *details);
static void sd_bdii_freeVOList(SDVOList *vos);
static void sd_bdii_freeServiceDetailsList(SDServiceDetailsList *servicedetailsList);

/* Constructor/destructor for automatic plugin registration */
void SD_initPlugin_bdii(void) __attribute__((__constructor__));
void SD_donePlugin_bdii(void) __attribute__((__destructor__));


/**********************************************************************
 * Global variables
 */

/* Service discovery plugin descriptor */
const SDPlugin SDPlugin_bdii =
{
	.name = "bdii",

	.getServiceDetails = sd_bdii_getServiceDetails,
	.listAssociatedServices = sd_bdii_listAssociatedServices,
	.listServicesByData = sd_bdii_listServicesByData,

	/* We also implement some non-mandatory methods since they can be
	 * optimized */
	.listServices = NULL,
	.listServicesByHost = sd_bdii_listServicesByHost,
	.getService = sd_bdii_getService,
	.getServiceData = sd_bdii_getServiceData,
	.getServiceDataItem = sd_bdii_getServiceDataItem,
	.getServiceSite = sd_bdii_getServiceSite,
	.getServiceWSDL = sd_bdii_getServiceWSDL,

	.freeService = sd_bdii_freeService,
	.freeServiceList = sd_bdii_freeServiceList,
	.freeServiceDataList = sd_bdii_freeServiceDataList,
	.freeServiceDetails = sd_bdii_freeServiceDetails,
	.freeVOList = sd_bdii_freeVOList
};

/* XXX This should be configurable */
static const char *basedn = "o=grid";

/* Cached connection handle */
static LDAP *ldcache;
/* Use count */
static int ldcnt;

/* Data key object attributes */
static char *data_attrs[] =
{
	ATTR_DATAKEY,
	ATTR_DATAVALUE,
	NULL
};

static char *all_attrs[] =
{
	LDAP_ALL_USER_ATTRIBUTES,
	NULL
};

static char *no_attrs[] =
{
	LDAP_NO_ATTRS,
	NULL
};

static char *link_attrs[] =
{
	ATTR_LINK,
	NULL
};

static char *service_attrs[] =
{
	ATTR_NAME,
	ATTR_TYPE,
	ATTR_ENDPOINT,
	ATTR_VERSION,
	NULL
};

static char *site_attrs[] =
{
	ATTR_LINK,
	NULL
};

static char *sa_attrs[] =
{
	ATTR_SA_ROOT,
	ATTR_SA_ACBR,
	NULL
};

static char *wsdl_attrs[] =
{
	ATTR_WSDL,
	NULL
};


/**********************************************************************
 * Utility functions
 */

/* Return an LDAP connection to the BDII server */
static LDAP *get_connection(SDException *exception)
{
	char *lcg_infosys;
	int ret;

	if (ldcache)
	{
		ldcnt++;
		return ldcache;
	}

	lcg_infosys = getenv("LCG_GFAL_INFOSYS");
	if (!lcg_infosys)
	{
		SD_I_setException(exception, SDStatus_FAILURE, "LCG_GFAL_INFOSYS is not set");
		return NULL;
	}

	/* ldap_init() can handle the "host:port" syntax just fine */
	ldcache = ldap_init(lcg_infosys, LDAP_PORT);
	if (!ldcache)
	{
		char buf[256];

		snprintf(buf, sizeof(buf), "ldap_init() failed: %s", strerror(errno));
		SD_I_setException(exception, SDStatus_FAILURE, buf);
		return NULL;
	}

	ret = ldap_simple_bind_s(ldcache, "", "");
	if (ret != LDAP_SUCCESS)
	{
		char buf[256];

		snprintf(buf, sizeof(buf), "ldap_simple_bind_s() failed: %s", strerror(errno));
		SD_I_setException(exception, SDStatus_FAILURE, buf);
		ldap_unbind_s(ldcache);
		ldcache = NULL;
		return NULL;
	}

	ldcnt++;
	return ldcache;
}

static void put_connection(LDAP *ld)
{
	if (--ldcnt)
		return;
	ldap_unbind_s(ld);
	ldcache = NULL;
}

static void report_query_error(const char *fmt, LDAP *ld, LDAPMessage *msg,
	int err, SDException *exception)
{
	char *error;
	int ret;
        if(NULL != msg){
	    ret = ldap_result2error(ld, msg, TRUE);
	} else {
            ret = err;
        }
        error = g_strdup_printf(fmt, ldap_err2string(ret));
	SD_I_setException(exception, SDStatus_FAILURE, error);
	g_free(error);
}

#if 0
static void report_ldap_error(const char *fmt, LDAP *ld, SDException *exception)
{
	char *error;
	int ret;

	ldap_get_option(ld, LDAP_OPT_ERROR_NUMBER, &ret);
	error = g_strdup_printf(fmt, ldap_err2string(ret));
	SD_I_setException(exception, SDStatus_FAILURE, error);
	g_free(error);
}
#endif

/* Get the value of a single-valued attribute */
static char *get_single_value(LDAP *ld, LDAPMessage *msg, const char *attr)
{
	char **values, *res;

	values = ldap_get_values(ld, msg, attr);
	if (!values)
		return NULL;
	res = g_strdup(values[0]);
	ldap_value_free(values);
	return res;
}

static int parse_service_attrs(SDService *service, LDAP *ld, LDAPMessage *msg, SDException *exception)
{

	service->name = get_single_value(ld, msg, ATTR_NAME);
	service->type = get_single_value(ld, msg, ATTR_TYPE);
	service->endpoint = get_single_value(ld, msg, ATTR_ENDPOINT);
	service->version = get_single_value(ld, msg, ATTR_VERSION);


	if(NULL == service->version) {
	    char *major = get_single_value(ld,msg,ATTR_VERSION_MAJOR);
	    char *minor = get_single_value(ld,msg,ATTR_VERSION_MINOR);
	    char *patch = get_single_value(ld,msg,ATTR_VERSION_PATCH);
	    service->version = (char *)malloc(20*sizeof(char));
	    if(major && minor && patch) {
		sprintf(service->version,"%s.%s.%s",major,minor,patch);
		g_free(major);
		g_free(minor);
		g_free(patch);
	    } else if(major) {
		sprintf(service->version,"%s",major);
		g_free(major);
	    } else {
		sprintf(service->version,"N/A");
	    }
	}

	if(NULL == service->endpoint) {
	    service->endpoint = get_single_value(ld, msg, ATTR_URI);
	}

	if(NULL == service->name) {
	    service->name = get_single_value(ld, msg, ATTR_NM);
	    if(NULL == service->name) {
		service->name = get_single_value(ld, msg, ATTR_URI);
	    }
	}

	if (!service->name || !service->type || !service->endpoint)
	{
	    char msg[1000];
	    sprintf(msg,"Incomplete service definition: ServiceName:%s ServiceType:%s ServiceEndpoint:%s ServiceVersion:%s",
		    !service->name ? "N/A" : service->name ,
		    !service->type ? "N/A" : service->type ,
		    !service->endpoint ? "N/A" : service->endpoint ,
		    !service->version ? "N/A" : service->version );
		SD_I_setException(exception, SDStatus_FAILURE, msg);
		return -1;
	}

	// the SRM fix (you can start laughing now)
	if(strcmp(service->type,"srm_v1") == 0) {
	    strcpy(service->type,"SRM");
	}
	return 0;
}

static int match_vos(LDAP *ld, LDAPMessage *msg, const SDVOList *vos)
{
       char **values;
       int i,j;

       // try the VO as we know it (owner)
       values = ldap_get_values(ld, msg, ATTR_VO);
       if(!values) {
	   // try the VO as it is sometimes used (accessrule)
	   values = ldap_get_values(ld, msg, ATTR_ACCESSRULE);
       }

       // if no VO is given at all, return TRUE, assume that the service is for all VOs
       if(!values)
	   return TRUE;

       for(i = 0; i < vos->numNames; i++) {
	   for(j = 0; j < ldap_count_values(values); j++) {
	       if( 0 == strcasecmp(vos->names[i],values[j]) ) {
		   ldap_value_free(values);
		   return TRUE;
	       }
	   }
       }

       ldap_value_free(values);
       return FALSE;
}

static int fill_service(const char *name, SDService *service, SDException *exception)
{
	struct timeval timeout;
	LDAPMessage *reply;
	char *query;
	LDAP *ld;
	int ret;

	ld = get_connection(exception);
	if (!ld)
		return -1;

	query = g_strdup_printf(QUERY_NAME, name, name);

	timeout.tv_sec = QUERY_TIMEOUT;
	timeout.tv_usec = 0;

	ret = ldap_search_st(ld, basedn, LDAP_SCOPE_SUBTREE, query, all_attrs, 0, &timeout, &reply);
	g_free(query);
	if (ret != LDAP_SUCCESS)
	{
		report_query_error("LDAP search failed: %s", ld, reply, ret,exception);
		ldap_msgfree(reply);
		put_connection(ld);
		return -1;
	}

	ret = parse_service_attrs(service, ld, reply, exception);
	ldap_msgfree(reply);
	put_connection(ld);

	return ret;
}

static int add_associated_service(SDServiceDetails *details, const char *name, SDException *exception)
{
	SDService *service, **tmp;
	int ret;

	service = g_new0(SDService, 1);
	ret = fill_service(name, service, exception);
	if (ret)
	{
		/* We do not want to report an error if the pointed service does not exist */
		sd_bdii_freeService(service);
		return 0;
	}

	if (!details->associatedServices)
		details->associatedServices = g_new0(SDServiceList, 1);

	tmp = g_renew(SDService *, details->associatedServices->services,
		details->associatedServices->numServices + 1);
	tmp[details->associatedServices->numServices++] = service;
	details->associatedServices->services = tmp;
	return 0;
}

static int add_vo(SDServiceDetails *details, const char *vo)
{
	if (!details->vos)
		details->vos = g_new0(SDVOList, 1);

	details->vos->names = g_renew(char *, details->vos->names, details->vos->numNames + 1);
	details->vos->names[details->vos->numNames++] = g_strdup(vo);
	return 0;
}

/* Add a service to a SDServiceList */
static int add_to_list(SDServiceList *list, SDService *service, SDException *ex)
{
        SDService **tmp;

        tmp = g_renew(SDService *, list->services, list->numServices + 1);
        if (!tmp)
        {
		SD_I_setException(ex, SDStatus_FAILURE, "Out of memory");
		return -1;
        }

        tmp[list->numServices++] = service;
        list->services = tmp;
        return 0;
}

static int parse_detail_attrs(SDServiceDetails *details, LDAP *ld, LDAPMessage *msg,
	SDException *exception)
{
	char **values;
	int i, ret;

	SDService *service;
	service = g_new0(SDService, 1);
	ret = parse_service_attrs(service, ld, msg, exception);
	if(ret)
	    return ret;

	details->name = service->name;
	details->type = service->type;
	details->endpoint = service->endpoint;
	details->version = service->version;
	details->wsdl = get_single_value(ld, msg, ATTR_WSDL);
	/* details->administration */

	/* ->site and ->associatedServices comes from ATTR_LINK */
	values = ldap_get_values(ld, msg, ATTR_LINK);
	ret = 0;
	for (i = 0; values && values[i]; i++)
	{
		char *name, *val;

		name = g_strdup(values[i]);
		val = strchr(name, '=');
		if (!val)
		{
			g_free(name);
			continue;
		}

		*val++ = '\0';

		if (!strcasecmp(name, ATTR_SITE))
		{
			g_free(details->site);
			details->site = g_strdup(val);
		}
		else if (!strcasecmp(name, ATTR_NAME))
		{
			ret = add_associated_service(details, val, exception);
			if (ret)
				break;
		}
		g_free(name);
	}
	ldap_value_free(values);
	if (ret)
		return ret;


	if(!details->site) {
	    // it can be that the site is attached differently (old LCG model)
	    details->site = get_single_value(ld, msg, ATTR_HOSTINGORG);
	}

	values = ldap_get_values(ld, msg, ATTR_ACCESSRULE);
	if(!values) {
	    // try the VO as it should be used
	    values = ldap_get_values(ld, msg, ATTR_VO);
	}
	for (i = 0; values && values[i]; i++)
	{
		ret = add_vo(details, values[i]);
		if (ret)
			break;
	}
	ldap_value_free(values);
	if (ret)
		return ret;

	SD_I_setException(exception, SDStatus_SUCCESS, "");

	return 0;
}

static char *get_host(const char *endpoint) {

    char * ret = g_strdup(endpoint);

    char *start = strstr(ret, "://");
    if(NULL != start) {
	ret = start+3;
    }

    char *end = strchr(ret,':');
    if(NULL != end) {
	*end = '\0';
    } else {
	end = strchr(ret,'/');
	if(NULL != end)
	    *end = '\0';
    }

    return ret;

}

static void add_sa_root(SDServiceDataList *list, const char *endpoint, SDException *exception)
{
        char *host = get_host(endpoint);
	LDAPMessage *reply, *msg;
	LDAP *ld;
	struct timeval timeout;
	GString *query;
	int ret;

	ld = get_connection(exception);
	if (!ld)
	    return;

	timeout.tv_sec = QUERY_TIMEOUT;
	timeout.tv_usec = 0;

	query = g_string_new("(&" QUERY_OC_SA);
	g_string_append_printf(query, QUERY_SE, host);
	g_string_append(query, ")");

	ret = ldap_search_st(ld, basedn, LDAP_SCOPE_SUBTREE, query->str, sa_attrs, 0, &timeout, &reply);
	g_string_free(query, TRUE);
	if (ret != LDAP_SUCCESS)
	    {
		report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
		goto out;
	    }

	if(ldap_count_entries(ld, reply) == 0) {
	    goto out;
	}


	for (msg = ldap_first_entry(ld, reply); msg; msg = ldap_next_entry(ld, msg))
	    {
		char *vo = get_single_value(ld, msg, ATTR_SA_ACBR);
		char *root = get_single_value(ld, msg, ATTR_SA_ROOT);

		if(!vo || !root)
		    continue;
		    
		char *value = strchr(root,':');
		char *key = g_strdup_printf("%s:SEMountPoint",vo);
		g_free(vo);

		list->items = g_renew(SDServiceData, list->items, list->numItems + 1);
		list->items[list->numItems].key = key;

		if( NULL != value ) {
		    list->items[list->numItems].value = ++value;
		} else {
		    list->items[list->numItems].value = root;
		}
		++list->numItems;
	    }

 out:
	ldap_msgfree(reply);
	put_connection(ld);
	return;

}

static SDServiceDataList *query_service_data(LDAP *ld, LDAPMessage *imsg, 
					     const char *type, const char *endpoint, 
					     SDException *exception)
{

        char *dn;
	SDServiceDataList *list = NULL;
	LDAPMessage *reply, *msg;
	struct timeval timeout;
	int ret;

	dn = ldap_get_dn(ld, imsg);
	if (!dn)
	    return NULL;

	timeout.tv_sec = QUERY_TIMEOUT;
	timeout.tv_usec = 0;
	ret = ldap_search_st(ld, dn, LDAP_SCOPE_SUBTREE, QUERY_DATA, data_attrs, 0, &timeout, &reply);
	ldap_memfree(dn);
	if (ret != LDAP_SUCCESS)
	{
		report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
		goto error;
	}


	list = g_new0(SDServiceDataList, 1);
	for (msg = ldap_first_entry(ld, reply); msg; msg = ldap_next_entry(ld, msg))
	{
		char *key, *value;

		key = get_single_value(ld, msg, ATTR_DATAKEY);
		value = get_single_value(ld, msg, ATTR_DATAVALUE);
		if (!key || !value)
		{
			g_free(key);
			g_free(value);
			continue;
		}

		list->items = g_renew(SDServiceData, list->items, list->numItems + 1);
		list->items[list->numItems].key = key;
		list->items[list->numItems].value = value;
		++list->numItems;
	}

	ldap_msgfree(reply);

	// SPECIAL case for the SE/SRM case
	if( strncasecmp(type,"srm",3) == 0) {
	    add_sa_root(list, endpoint, exception);
	}

	return list;

error:
	sd_bdii_freeServiceDataList(list);
	ldap_msgfree(reply);
	return NULL;
}

static int match_data(const SDServiceDataList *list, const SDServiceDataList *requirements)
{
	int i, j, gotit;

	/* Check that the service has all the specified key/value
	 * pairs */
	for (i = 0; i < requirements->numItems; i++)
	{
		gotit = 0;

		for (j = 0; list && j < list->numItems; j++)
		{
			if (strcmp(requirements->items[i].key, list->items[j].key))
				continue;
			if (!strcmp(requirements->items[i].value, list->items[j].value))
			{
				gotit = 1;
				break;
			}
		}

		if (!gotit)
			return FALSE;
	}

	return TRUE;
}

/* Returns TRUE if the service matches the given constraints, FALSE otherwise */
static int match_service(SDServiceDetails *details, const char *type,
	const char *site, const SDVOList *vos)
{
	int i, j, gotit;

	if (type && strcmp(details->type, type))
		return FALSE;
	if (site && details->site && strcmp(details->site, site))
		return FALSE;

	/* <volist> is not a required parameter. Treat an missing <volist>
	 * as a "match all", an empty <volist> as a "match none" */
	if (vos && details->vos)
	{
		gotit = 0;

		/* A service that does not belong to any VOs matches the
		 * empty list */
		if (!vos->numNames && !details->vos->numNames)
			gotit = 1;

		/* Check that at the intersection of vos and details->vos
		 * is not empty */
		for (i = 0; i < vos->numNames && !gotit; i++)
		{
			for (j = 0; details->vos && j < details->vos->numNames;
					j++)
			{
				if (!strcmp(vos->names[i],
						details->vos->names[j]))
				{
					gotit = 1;
					break;
				}
			}
		}

		if (!gotit)
			return FALSE;
	}

	return TRUE;
}

/* Convert SDServiceDetails to SDService */
static SDService *details_to_service(const SDServiceDetails *details)
{
	SDService *service;

	service = g_new0(SDService, 1);
	service->name = g_strdup(details->name);
	service->type = g_strdup(details->type);
	service->endpoint = g_strdup(details->endpoint);
	service->version = g_strdup(details->version);
	return service;
}

static int get_details(const char *q, const char *name, LDAP *ld, LDAPMessage **reply, SDException *exception) 
{
        char *query;
	struct timeval timeout;
	int ret;

	query = g_strdup_printf(q, name, name);

	timeout.tv_sec = QUERY_TIMEOUT;
	timeout.tv_usec = 0;

	ret = ldap_search_st(ld, basedn, LDAP_SCOPE_SUBTREE, query, all_attrs, 0, &timeout, reply);
	g_free(query);
	if (ret != LDAP_SUCCESS)
	{
		report_query_error("LDAP search failed: %s", ld, *reply, ret, exception);
		return -1;
	}
	return 0;
}

static SDServiceList *do_list_query(const SDServiceDataList *data,
				    const SDVOList *vos, LDAP *ld, GString *query,
				    SDException *exception)
{
	LDAPMessage *reply, *msg;
	struct timeval timeout;
	SDServiceList *list;
	int ret;

	timeout.tv_sec = QUERY_TIMEOUT;
	timeout.tv_usec = 0;

	ret = ldap_search_st(ld, basedn, LDAP_SCOPE_SUBTREE, query->str, all_attrs, 0, &timeout, &reply);
	g_string_free(query, TRUE);
	if (ret != LDAP_SUCCESS)
	{
		report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
		list = NULL;
		goto out;
	}

	if(ldap_count_entries(ld, reply) == 0) {
	    SD_I_setException(exception, SDStatus_FAILURE, "No entries found");
	    list = NULL;
	    goto out;
	}


	list = g_new0(SDServiceList, 1);
	for (msg = ldap_first_entry(ld, reply); msg; msg = ldap_next_entry(ld, msg))
	{

		SDService *service;

		service = g_new0(SDService, 1);
		ret = parse_service_attrs(service, ld, msg, exception);
		if (ret)
		{
			sd_bdii_freeService(service);
			sd_bdii_freeServiceList(list);
			list = NULL;
			goto out;
		}
		
		if (data)
		{
			SDServiceDataList *datalist;

			SD_I_setException(exception, SDStatus_SUCCESS, "");
			datalist = query_service_data(ld, msg, service->type, service->endpoint, exception);
			if (exception->status != SDStatus_SUCCESS)
				continue;


			/* Filter out non-matching entries */
			if (!match_data(datalist, data))
				continue;
		}

		if(vos)
                {
		        /* Filter out entries not matching the requested VOs */
		        if(!match_vos(ld, msg, vos))
			        continue;
		}

		ret = add_to_list(list, service, exception);
		if (ret)
		{
			sd_bdii_freeService(service);
			sd_bdii_freeServiceList(list);
			list = NULL;
			goto out;
		}
	}

	SD_I_setException(exception, SDStatus_SUCCESS, "");
out:
	ldap_msgfree(reply);
	put_connection(ld);
	return list;
}


/**********************************************************************
 * Service discovery plugin implementation
 */

static SDServiceDetails *sd_bdii_getServiceDetails(const char *name,
						   SDException *exception)
{
	SDServiceDetails *details = NULL;
	LDAPMessage *reply;
	LDAP *ld;
	int ret;

	ld = get_connection(exception);
	if (!ld)
		return NULL;

	/* query_service_data() returns the error code in the exception */
	if (!exception)
		exception = g_newa(SDException, 1);

	ret = get_details(QUERY_NAME, name, ld, &reply, exception);
	if( ret < 0 )
	    goto error;


	// nothing?
	if(ldap_count_entries(ld, reply) == 0) {
	    SD_I_setException(exception, SDStatus_FAILURE, "No entry found");
	    goto error;
	}

	details = g_new0(SDServiceDetails, 1);
	ret = parse_detail_attrs(details, ld, reply, exception);
	if (ret)
		goto error;

	SD_I_setException(exception, SDStatus_SUCCESS, "");
	details->data = query_service_data(ld, reply, details->type, details->endpoint, exception);
	if (exception->status != SDStatus_SUCCESS)
		goto error;

out:
	ldap_msgfree(reply);
	put_connection(ld);
	return details;

error:
	if (details)
		sd_bdii_freeServiceDetails(details);
	details = NULL;
	goto out;
}

static SDServiceList *sd_bdii_listAssociatedServices(const char *name,
	const char *type, const char *site, const SDVOList *vos,
	SDException *exception)
{
	struct timeval timeout;
	char *query, **values;
	SDServiceList *list;
	SDServiceDetailsList *detailslist;
	LDAPMessage *reply;
	int ret, i;
	LDAP *ld;

	ld = get_connection(exception);
	if (!ld)
		return NULL;

	query = g_strdup_printf(QUERY_NAME, name, name);

	timeout.tv_sec = QUERY_TIMEOUT;
	timeout.tv_usec = 0;

	ret = ldap_search_st(ld, basedn, LDAP_SCOPE_SUBTREE, query, link_attrs, 0, &timeout, &reply);
	g_free(query);
	if (ret != LDAP_SUCCESS)
	{
		report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
		return NULL;
	}

	list = g_new0(SDServiceList, 1);
	detailslist = g_new0(SDServiceDetailsList, 1);
	if (!list || !detailslist)
	{
		SD_I_setException(exception, SDStatus_FAILURE, "Out of memory");
		return NULL;
	}

	values = ldap_get_values(ld, reply, ATTR_LINK);
	for (i = 0; values && values[i]; i++)
	{
		SDServiceDetails *details;
                SDService *service;
		int ret;

		/* We are only interested in service links */
		if (strncasecmp(values[i], ATTR_NAME, strlen(ATTR_NAME)) ||
				values[i][strlen(ATTR_NAME)] != '=')
			continue;

		/* A non-existant associated service is not an error */
		details = sd_bdii_getServiceDetails(values[i] + strlen(ATTR_NAME) + 1, exception);
		if (!details)
			continue;

		/* Check if the service matches the requirements */
		if (!match_service(details, type, site, vos))
		{
			sd_bdii_freeServiceDetails(details);
			continue;
		}

                service = details_to_service(details);
		ret = add_to_list(list, service, exception);
		if (ret)
		{
			sd_bdii_freeService(service);
			sd_bdii_freeServiceList(list);
			sd_bdii_freeServiceDetails(details);
			list = NULL;
			goto out;
		}
	}
	
	if (!list->numServices)
	{
		char buf[256];
		sd_bdii_freeServiceList(list);
		snprintf(buf, 255, "No services of type %s were found", type);
		SD_I_setException(exception, SDStatus_FAILURE, buf);
		list = NULL;
		goto out;
	}

	SD_I_setException(exception, SDStatus_SUCCESS, "");

out:
	ldap_value_free(values);
	ldap_msgfree(reply);
	put_connection(ld);
	return list;
}

static SDServiceList *sd_bdii_listServicesByHost(const char *type, const char *host, const SDVOList *vos,
						 SDException *exception)
{
	GString *query;
	LDAP *ld;

        if(host == NULL)
	        return sd_bdii_listServicesByData(NULL,type,NULL,vos,exception);


	ld = get_connection(exception);
	if (!ld)
		return NULL;

	/* do_list_query returns the error code in the exception */
	if (!exception)
		exception = g_newa(SDException, 1);

	/* Set up a general LDAP error message */
	SD_I_setException(exception, SDStatus_FAILURE, "LDAP query failed");

	/* GString is _very_ handy here :-) */
	query = g_string_new("(&" QUERY_OC_SERVICE);

    if (type) {
        // this is the SRM hack so that both old and new types are looked for
        if(strncasecmp(type,"srm",3) == 0) {
            g_string_append(query, QUERY_SRM_TYPE);
        } else {
            g_string_append_printf(query, QUERY_TYPE, type);
        }
    }

	g_string_append_printf(query, QUERY_HOST, host, host);

	/* closing the and-ed subqueries */
	g_string_append(query, ")");

	return do_list_query(NULL,vos,ld,query,exception);

}


static SDServiceList *sd_bdii_listServicesByData(const SDServiceDataList *data,
						 const char *type, const char *site, const SDVOList *vos,
						 SDException *exception)
{
	/* type, site and vos can be expressed as an LDAP filter. Data
	 * key-value pairs however appear as child nodes in BDII so we have
	 * to check them manually. */
	GString *query;
	LDAP *ld;

	ld = get_connection(exception);
	if (!ld)
		return NULL;

	/* do_list_query returns the error code in the exception */
	if (!exception)
		exception = g_newa(SDException, 1);

	/* Set up a general LDAP error message */
	SD_I_setException(exception, SDStatus_FAILURE, "LDAP query failed");

	/* GString is _very_ handy here :-) */
	if(type || site)
	    query = g_string_new("(&" QUERY_OC_SERVICE);
	else
	    query = g_string_new(QUERY_OC_SERVICE);

	if (type) {
	    // this is the SRM hack so that both old and new types are looked for
	    if(strncasecmp(type,"srm",3) == 0) {
		g_string_append(query, QUERY_SRM_TYPE);
	    } else {
		g_string_append_printf(query, QUERY_TYPE, type);
	    }
	}
	if (site)
		g_string_append_printf(query, QUERY_SITE, site);


	/* closing the and-ed subqueries */
	if(type || site)
	    g_string_append(query, ")");


	return do_list_query(data,vos,ld,query,exception);
}

static SDService *sd_bdii_getService(const char *name, SDException *exception)
{

	SDService *service = NULL;
	GString *query = NULL;
	LDAP *ld;

	ld = get_connection(exception);
	if (!ld)
		return NULL;

	/* do_list_query returns the error code in the exception */
	if (!exception)
		exception = g_newa(SDException, 1);

	/* Set up a general LDAP error message */
	SD_I_setException(exception, SDStatus_FAILURE, "LDAP query failed");

	/* GString is _very_ handy here :-) */
	if(!name) {
	    SD_I_setException(exception, SDStatus_FAILURE, "Name not given in Service search");
	    return NULL;
	}

	query = g_string_sized_new(200);
	g_string_printf(query, QUERY_NAME, name, name);

	SDServiceList *list = do_list_query(NULL, NULL, ld, query, exception);
    if(list == NULL)
        return NULL;
	if(list->numServices < 1)
	    return NULL;

	service = list->services[0];
	g_free(list->services);
	g_free(list);

	return service;
}

static SDServiceDataList *sd_bdii_getServiceData(const char *name,
	SDException *exception)
{
	SDServiceDataList *list;
	struct timeval timeout;
	LDAPMessage *reply;
	char *query;
	LDAP *ld;
	int ret;

	ld = get_connection(exception);
	if (!ld)
		return NULL;

	/* query_service_data() returns the error code in the exception */
	if (!exception)
		exception = g_newa(SDException, 1);

	query = g_strdup_printf(QUERY_NAME, name, name);

	timeout.tv_sec = QUERY_TIMEOUT;
	timeout.tv_usec = 0;

	ret = ldap_search_st(ld, basedn, LDAP_SCOPE_SUBTREE, query, service_attrs, 0, &timeout, &reply);
	g_free(query);
	if (ret != LDAP_SUCCESS)
	{
		report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
		list = NULL;
		goto out;
	}

	SDService *service;

	service = g_new0(SDService, 1);
	ret = parse_service_attrs(service, ld, reply, exception);
	if (ret)
	{
		sd_bdii_freeService(service);
		list = NULL;
		goto out;
	}

	SD_I_setException(exception, SDStatus_SUCCESS, "");
	list = query_service_data(ld, reply, service->type, service->endpoint, exception);

out:
	ldap_msgfree(reply);
	put_connection(ld);
	return list;
}

static char *sd_bdii_getServiceSite(const char *name, SDException *exception)
{
	char *query, **values, *res = NULL;
	struct timeval timeout;
	LDAPMessage *reply;
	int ret, i;
	LDAP *ld;

	ld = get_connection(exception);
	if (!ld)
		return NULL;

	query = g_strdup_printf(QUERY_NAME, name, name);

	timeout.tv_sec = QUERY_TIMEOUT;
	timeout.tv_usec = 0;

	ret = ldap_search_st(ld, basedn, LDAP_SCOPE_SUBTREE, query, site_attrs, 0, &timeout, &reply);
	g_free(query);
	if (ret != LDAP_SUCCESS)
	{
		report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
		goto out;
	}

	values = ldap_get_values(ld, reply, ATTR_LINK);
	for (i = 0; values && values[i]; i++)
	{
		if (strncasecmp(values[i], ATTR_SITE, strlen(ATTR_SITE)) ||
				values[i][strlen(ATTR_SITE)] != '=')
			continue;
		res = strdup(values[i] + strlen(ATTR_SITE) + 1);
	}
	ldap_value_free(values);

out:
	ldap_msgfree(reply);
	put_connection(ld);
	return res;
}

static char *sd_bdii_getServiceWSDL(const char *name, SDException *exception)
{
	char *query, *res = NULL;
	struct timeval timeout;
	LDAPMessage *reply;
	LDAP *ld;
	int ret;

	ld = get_connection(exception);
	if (!ld)
		return NULL;

	query = g_strdup_printf(QUERY_NAME, name, name);

	timeout.tv_sec = QUERY_TIMEOUT;
	timeout.tv_usec = 0;

	ret = ldap_search_st(ld, basedn, LDAP_SCOPE_SUBTREE, query, wsdl_attrs, 0, &timeout, &reply);
	g_free(query);
	if (ret != LDAP_SUCCESS)
	{
		report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
		goto out;
	}

	res = get_single_value(ld, reply, ATTR_WSDL);
	if (!g_mem_is_system_malloc() && res)
	{
		char *tmp = strdup(res);
		g_free(res);
		res = tmp;
	}

	SD_I_setException(exception, SDStatus_SUCCESS, "");
out:
	ldap_msgfree(reply);
	put_connection(ld);
	return res;
}

static char *sd_bdii_getServiceDataItem(const char *name, const char *key, SDException *exception)
{
	char *query, *dn, *res = NULL;
	struct timeval timeout;
	LDAPMessage *reply;
	LDAP *ld;
	int ret;

	ld = get_connection(exception);
	if (!ld)
		return NULL;

	query = g_strdup_printf(QUERY_NAME, name, name);

	timeout.tv_sec = QUERY_TIMEOUT;
	timeout.tv_usec = 0;

	ret = ldap_search_st(ld, basedn, LDAP_SCOPE_SUBTREE, query, no_attrs, 0, &timeout, &reply);
	g_free(query);
	if (ret != LDAP_SUCCESS)
	{
		report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
		goto out;
	}

	dn = ldap_get_dn(ld, reply);
	if (!dn)
	{
		SD_I_setException(exception, SDStatus_FAILURE, "Failed to determine the DN");
		goto out;
	}

	ldap_msgfree(reply);

	query = g_strdup_printf(QUERY_DATAKEY, key);
	timeout.tv_sec = QUERY_TIMEOUT;
	timeout.tv_usec = 0;

	ret = ldap_search_st(ld, dn, LDAP_SCOPE_SUBTREE, query, data_attrs, 0, &timeout, &reply);
	g_free(query);
	ldap_memfree(dn);
	if (ret != LDAP_SUCCESS)
	{
		report_query_error("LDAP search failed: %s", ld, reply, ret, exception);
		goto out;
	}

	res = get_single_value(ld, reply, ATTR_DATAVALUE);
	if (!g_mem_is_system_malloc() && res)
	{
		char *tmp = strdup(res);
		g_free(res);
		res = tmp;
	}

	SD_I_setException(exception, SDStatus_SUCCESS, "");
out:
	ldap_msgfree(reply);
	put_connection(ld);
	return res;
}

/**********************************************************************
 * Memory management methods
 */

static void sd_bdii_freeService(SDService *service)
{
	if (!service)
		return;

	g_free(service->name);
	g_free(service->type);
	g_free(service->endpoint);
	g_free(service->version);
	g_free(service);
}

static void sd_bdii_freeServiceList(SDServiceList *list)
{
	int i;

	if (!list)
		return;

	for (i = 0; i < list->numServices; i++)
		sd_bdii_freeService(list->services[i]);
	g_free(list->services);
	g_free(list);
}

static void sd_bdii_freeServiceDataList(SDServiceDataList *list)
{
	int i;

	if (!list)
		return;
	for (i = 0; i < list->numItems; i++)
	{
		g_free(list->items[i].key);
		g_free(list->items[i].value);
	}
	g_free(list->items);
	g_free(list);
}

static void sd_bdii_freeServiceDetails(SDServiceDetails *details)
{
	if (!details)
		return;

	g_free(details->name);
	g_free(details->type);
	g_free(details->endpoint);
	g_free(details->version);
	g_free(details->site);
	g_free(details->wsdl);
	g_free(details->administration);
	sd_bdii_freeVOList(details->vos);
	sd_bdii_freeServiceList(details->associatedServices);
	sd_bdii_freeServiceDataList(details->data);
	g_free(details);
}

static void sd_bdii_freeServiceDetailsList(SDServiceDetailsList *servicedetailsList)
{
	int i;
        if (!servicedetailsList)
                return;

	for (i = 0; i < servicedetailsList->numServiceDetails; i++)
		g_free(servicedetailsList->servicedetails[i]);
        g_free(servicedetailsList->servicedetails);
        g_free(servicedetailsList);
}

static void sd_bdii_freeVOList(SDVOList *vos)
{
	int i;

	if (!vos)
		return;
	for (i = 0; i < vos->numNames; i++)
		g_free(vos->names[i]);
	g_free(vos->names);
	g_free(vos);
}


/**********************************************************************
 * Service discovery plugin registration
 */

void SD_initPlugin_bdii(void)
{
	SD_registerPlugin(&SDPlugin_bdii, NULL);
}

void SD_donePlugin_bdii(void)
{
	SD_unregisterPlugin(&SDPlugin_bdii);
}
